;;;; -*- Mode:Lisp; Package:CL-USER; Syntax:COMMON-LISP; Base:10 -*-

(in-package :SFS)

;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;; Copyright (c) 1996 by Micheal Scott Hewett
;;;
;;; This code may be used by anyone for any project, but may not
;;; be sold in source or object form without permission.
;;; If in doubt, follow the GNU "copyleft" guidelines.
;;;
;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;; (Contact: hewett@cs.utexas.edu or hewett@cs.stanford.edu)
;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;;
;;;  A unifier for the SFS.
;;;
;;; 12 Nov 1996 (mh)  Copied the AAM unify function.
;;;                   Need to copy it for package independence.
;;;    




;;; ---------------  UNIFICATION  ------------------

;;; In these functions, 'pat' is something like: (color ?x    ?y)
;;; and 'exp' is something like:                 (color house blue)
;;; We want to return only bindings for variables in the 'pat', but we
;;; need to use variables in the 'exp' to do a correct unification,
;;; so we keep two binding lists around.

;;; Each input is a clause.

;;; A good complex test is: (sfs::unify '(r ?x ?z B) '(r B ?y ?y))
;;; which should return:    (T (?Z . B) (?X . B))

;;; another test:           (sfs::unify '(r ?x ?y ?x ?z) '(r a ?b ?b a))
;;; which should return:    (T (?Z . A) (?Y . A) (?X . A))

;;; test: (sfs::unify '(P P ?Y H ?Z) '(?F ?C ?C G ?D))  should fail

;;; test: (sfs::unify '(?W J ?X ?U L ?Z N ?X) '(?B J N ?B L P N P))   should fail


;;; For the SFS, we know that the expression is always ground,
;;; and the pattern is usually not.


(defun unify (pattern expression) 
  "Returns T or NIL."
  
  
  (monitor :SELECT)

  (let ((p-and-e-info (unify-aux pattern expression (list NIL NIL NIL NIL))))
      
    ;; Some variables in the pattern may be bound to variables
    ;; in the expression that now have bindings.  Sort those
    ;; out.
      
    (not (eq (car p-and-e-info) :FAILED))
    )
  )


(defun unify-aux (pattern expression p-and-e-bindings)

  (cond ((eq (car p-and-e-bindings) :FAILED)      p-and-e-bindings)

	((null pattern)     (if (null expression) p-and-e-bindings (cons :FAILED :FAILED)))

	((consp pattern)	(unify-aux
				 (cdr pattern) (cdr expression) 
				 (unify-aux (car pattern) (car expression) p-and-e-bindings)))

	(T

	 (unify-aux-2 pattern expression
			   (first p-and-e-bindings)
			   (second p-and-e-bindings)
			   (third p-and-e-bindings)
			   (fourth p-and-e-bindings)))
	)
  )


(defun unify-aux-2 (pattern expression p-bindings e-bindings p-classes e-classes)

  (let ((pat-type   NIL)
	(exp-type   NIL)
	)

    (setq pat-type (equivalence-type-of pattern    p-classes))
    (setq exp-type (equivalence-type-of expression e-classes))

    ;; If either one is a member of another equivalence class,
    ;; we need to do some equivalence class merging or creating.

    (cond ((and (eq pat-type :NEW) (eq exp-type :NEW))
	   (unify-aux-simple pattern expression p-bindings e-bindings p-classes e-classes))


	  ((and (eq pat-type :NEW) (eq exp-type :MEMBER))
	   
	   (setq p-classes
	     (make-equivalence pattern (car (equiv-class-of
					     (binding-of (primary-of expression e-classes)
							 e-bindings)
					     p-classes))
			       p-classes))

	   (when (not (variable-p pattern))
	     (setq e-classes
	       (make-equivalence (primary-of expression e-classes) pattern e-classes)))

	   (when (or (eq e-classes :FAILED)
		     (eq p-classes :FAILED))
	     (setq p-bindings :FAILED))

	   (list p-bindings e-bindings p-classes e-classes))

	  ((and (eq pat-type :MEMBER) (eq exp-type :NEW))
	   
	   (setq e-classes
	     (make-equivalence expression (car (equiv-class-of
						(binding-of (primary-of pattern p-classes)
							    p-bindings)
						e-classes))
			       e-classes))

	   (when (not (variable-p expression))
	     (setq p-classes
	       (make-equivalence (primary-of pattern p-classes) expression p-classes)))

	   (when (or (eq e-classes :FAILED)
		     (eq p-classes :FAILED))
	     (setq p-bindings :FAILED))
	   
	   (list p-bindings e-bindings p-classes e-classes))


	  (T  ;; pat-type and exp-type are both :MEMBER.

	   (unless (eq (binding-of (primary-of pattern p-classes) p-bindings)   ;; already known?
		       (primary-of expression e-classes))

	     (multiple-value-setq
		 (p-classes p-bindings)
	       (merge-classes (equiv-class-of pattern p-classes)
			      (equiv-class-of (binding-of (primary-of expression e-classes)
							  e-bindings)
					      p-classes)
			      p-classes p-bindings))

	     (if (eq p-classes :FAILED)
		 (setq p-bindings :FAILED)
	       ;;else
	       (multiple-value-setq
		   (e-classes e-bindings)
		 (merge-classes (equiv-class-of expression e-classes)
				(equiv-class-of (binding-of (primary-of pattern p-classes)
							    p-bindings)
						e-classes)
				e-classes e-bindings)))

	     (when (eq e-classes :FAILED)
	       (setq p-bindings :FAILED))
	     )

	   (list p-bindings e-bindings p-classes e-classes))
	  )
    )
  )
	   

(defun unify-aux-simple (pattern expression p-bindings e-bindings p-classes e-classes)
  "Handles simple situations where the pattern and expression are all new."

  (unless (and (not (variable-p pattern))
	       (not (variable-p expression)))

    ;; Make equivalence classes for these.
    (setq p-classes (cons (list pattern)    p-classes))
    (setq e-classes (cons (list expression) e-classes))
    )

  (cond ((or (variable-p pattern)
	     (variable-p expression))
	 (setq p-bindings (bind pattern expression p-bindings))
	 (setq e-bindings (bind expression pattern e-bindings))

	 (list p-bindings e-bindings p-classes e-classes)
	 )
            
	 ;; Names are strings, so handle them separately.  Two strings is okay for EQUAL below.
	 ((and (stringp pattern) (not (stringp expression)))
	  (if (funcall *equalp-test* (read-from-string pattern) expression)
	      (list p-bindings e-bindings p-classes e-classes)
	    ;;else
	    (list :FAILED e-bindings p-classes e-classes)))

	 ((and (stringp expression) (not (stringp pattern)))
	  (if (funcall *equalp-test* (read-from-string expression) pattern)
	      (list p-bindings e-bindings p-classes e-classes)
	    ;;else
	    (list :FAILED e-bindings p-classes e-classes)))

	 (T   (if (funcall *equalp-test* pattern expression)
		  (list p-bindings e-bindings p-classes e-classes)
		;;else
		(list :FAILED e-bindings p-classes e-classes)))
	 )
  )



(defun bind (var value bindings)

  ;; Puts (var . value) on the bindings list unless it finds
  ;; a binding for that variable already there.
  ;; If the existing binding is the same, that's fine.
  ;; If it's different, unification fails.

  (let ((entry (assoc var bindings)))

    (cond (entry  (if (eq value (cdr entry))
		      bindings :FAILED))

          (T      (cons (cons var value) bindings))
          )
    )
  )


(defun equivalence-type-of (literal classes)
  "Returns either :NEW, or :MEMBER.
Use MAKE-EQUIVALENCE to create a new one."

  ;; If 'literal' is a constant, it is always the primary member
  ;; of its equivalence class.
  
  (let ((class  (equiv-class-of literal classes)))

    (if (null class)
	:NEW
      ;;else
      :MEMBER
      )
    )
  )


(defun primary-of (literal classes)
  "It is guaranteed that literal is a member of one of the classes."

  (car (equiv-class-of literal classes)))


(defun binding-of (literal bindings)
  "Returns NIL if no binding exists."

  (cdr (assoc literal bindings)))


(defun equiv-class-of (literal classes)
  (find literal classes :test #'(lambda (lit class)
				  (member lit class :test #'equal))))
  

(defun make-equivalence (literal1 literal2 classes)

  (if (and (not (variable-p literal1))
	   (not (variable-p literal2)))

      (if (not (eq literal1 literal2))
	  :FAILED
	classes)
    ;;else

    (let ((class (or (find literal1 classes :test #'member)
		     (find literal2 classes :test #'member)))

	  (primary (if (variable-p literal1) literal2 literal1))
	  (other   (if (variable-p literal1) literal1 literal2))
	  )

      (cond ((null class)
	     (push (list primary other) classes))

	    ((member primary class)
	     (push other (cdr class)))

	    ((not (variable-p primary))   ;; it becomes the primary
	     (setq classes (delete class classes :test #'equal))
	     (push (push primary class) classes))

	    (T   (push primary (cdr class)))
	    )

      classes
      )
    )
  )



(defun merge-classes (class1 class2 classes bindings)
  "Merges the two and returns a new list of classes and a new list of bindings."

  ;; Make sure we aren't doing an illegal merge
  (when (and (not (variable-p (car class1)))
	     (not (variable-p (car class2)))
	     (not (eq (car class1) (car class2))))
    (return-from MERGE-CLASSES (values :FAILED :FAILED))
    )

  ;; Ensure that a constant always remains the primary member of the class.
  (when (and (variable-p (car class1)) (not (variable-p (car class2))))
    (psetq class1 class2
	   class2 class1))

  (unless (equal class1 class2)
    (setq classes  (delete (car class2) classes  :key #'car))
    (setq bindings (delete (car class2) bindings :key #'car))
    (nconc class1 class2)
    )

  (values classes bindings)
  )





(defun test-unify (&OPTIONAL (length 5))

  (let ((p-vars     '(?x ?y ?z ?u ?v ?w g h i j k l m n o p))
	(e-vars     '(?a ?b ?c g h i j k l m n o p))
	(pattern    NIL)
	(expression NIL)
	)

    (dotimes (i length)
      (push (nth (random (length p-vars)) p-vars) pattern)
      )

    (dolist (p-var pattern)
      (if (sfs::variable-p p-var)
	  (push (nth (random (length e-vars)) e-vars) expression)
	;;else
	  (push p-var expression)))

    (setq expression (nreverse expression))

    (format t "~%(sfs::unify '~S '~S)" pattern expression)

    (sfs::unify pattern expression)
    )
  )
